#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2011 Loic Jaquemet loic.jaquemet+python@gmail.com
#

__author__ = "Loic Jaquemet loic.jaquemet+python@gmail.com"

import ctypes
import logging

from ptrace.debugger.memory_mapping import readProcessMappings

from haystack.model import is_valid_address,is_valid_address_value,pointer2bytes,array2bytes,bytes2array,getaddress
from haystack.model import LoadableMembers,RangeValue,NotNull,CString,EVP_CIPHER_CTX_APP_DATA_PTR, IgnoreMember
from haystack import model
from ctypes_openssl import EVP_CIPHER_CTX, EVP_MD, HMAC_CTX, EVP_AES_KEY, AES_KEY,rijndael_ctx,EVP_RC4_KEY

log=logging.getLogger('openssh.model')


MODE_MAX=2 #kex.h:62
AES_BLOCK_LEN=16 #umac.c:168
AES_BLOCK_SIZE=16 
RIJNDAEL_BLOCKSIZE=16
L1_KEY_LEN=1024 #umac.c:298
L1_KEY_SHIFT=16 #umac.c:316
UMAC_OUTPUT_LEN=8 #umac.c:55
STREAMS=(UMAC_OUTPUT_LEN / 4) #umac.c:310
HASH_BUF_BYTES=64 # umac.c:315
SSH_SESSION_KEY_LENGTH=32 # ssh.h:84
MAXNR = 14

UINT64=ctypes.c_uint64
UINT32=ctypes.c_uint32
UINT8=ctypes.c_uint8



class OpenSSHStruct(LoadableMembers):
  ''' defines classRef '''
  pass

## Cipher's custom contexts
## cf EVP_CIPHER_CTX.app_data
class ssh_aes_ctr_ctx(OpenSSHStruct):
  ''' cipher-ctr.c:39 
  Cipher-custom context
  '''
  _fields_ = [
	('aes_ctx', AES_KEY),
	('aes_counter', ctypes.c_ubyte*AES_BLOCK_SIZE)
	]
  def getCtx(self):
    rd_key = self.aes_ctx.getKey()
    rounds = self.aes_ctx.getRounds()
    return (rd_key,rounds) 
  def getCounter(self):
    return array2bytes(self.aes_counter)
  def fromPyObj(self,pyobj):
    #recurse copy aes_ctx
    self.aes_ctx = AES_KEY().fromPyObj(pyobj.aes_ctx)
    #copy counter
    self.aes_counter=bytes2array(pyobj.aes_counter, ctypes.c_ubyte)
    return self



class rijndael_ctx(OpenSSHStruct):
  _fields_ = [
  ('decrypt', ctypes.c_int),
  ('Nr', ctypes.c_int),
  ('ek', ctypes.c_uint32*(4*(MAXNR + 1))),
  ('dk', ctypes.c_uint32*(4*(MAXNR + 1)))
  ]
  def getEk(self):
    return array2bytes(self.ek)
  def getDk(self):
    return array2bytes(self.dk)
  def fromPyObj(self,pyobj):
    self.decrypt = pyobj.decrypt
    self.Nr = pyobj.Nr
    self.ek = bytes2array(pyobj.ek, ctypes.c_uint32)
    self.dk = bytes2array(pyobj.dk, ctypes.c_uint32)
    return self
  
class ssh_rijndael_ctx(OpenSSHStruct):
  ''' cipher-aes.c:43 '''
  _fields_ = [
  ('r_ctx', rijndael_ctx),
  ('r_iv', ctypes.c_ubyte*RIJNDAEL_BLOCKSIZE)
  ]
  def getCtx(self):
    return self.r_ctx.getKey() 
  def getIV(self):
    return array2bytes(self.r_iv)
  def fromPyObj(self,pyobj):
    #recurse copy aes_ctx
    self.r_ctx = rijndael_ctx().fromPyObj(pyobj.r_ctx)
    #copy counter
    self.r_iv = bytes2array(pyobj.r_iv, ctypes.c_ubyte)
    return self


#EVP_CIPHER_CTX_APP_DATA_PTR._fields_=[ ('contents', ctypes.POINTER(ctypes.c_ubyte)) ,
#        ('ssh_aes_ctr_ctx',ctypes.POINTER(ssh_aes_ctr_ctx)) ]
    
class Cipher(OpenSSHStruct):
  ''' cipher.c:60 '''
  _fields_ = [
  ("name",  CString ), # yeah,  c_char_p can't handle us
  ("number",  ctypes.c_int), # for ssh1 only cipher.c and cipher.h:45 // -3 == SSH_CIPHER_SSH2
  ("block_size",  ctypes.c_uint), 
  ("key_len",  ctypes.c_uint), 
  ("discard_len",  ctypes.c_uint), 
  ("cbc_mode",  ctypes.c_uint), 
  ("evptype",  ctypes.POINTER(ctypes.c_int)) ## fn const EVP_CIPHER * function() qui renvoit & de la structure EVP_CIPHER
  ]
  expectedValues = {
  'name': NotNull,
  }
 

class CipherContext(OpenSSHStruct):
  ''' cipher.h:65 '''
  _fields_ = [
  ("plaintext",  ctypes.c_int), 
  ("evp",  EVP_CIPHER_CTX),
  ("cipher", ctypes.POINTER(Cipher))
  ]
  expectedValues = {
  'plaintext': [0,1],
  'cipher': NotNull
  }
  cipherContexts={ # we could check SSH_CIPHER_XXX in self.cipher.contents.number
	 "none": (None,None),
	 "des": (None,None),
	 "3des": (None,None),
	 "blowfish": (None,None),
	 "3des-cbc": (None,None),
	 "blowfish-cbc": (None,None),
	 "cast128-cbc": (None,None),
	 "arcfour": (EVP_RC4_KEY,'cipher_data'),
	 "arcfour128": (EVP_RC4_KEY,'cipher_data'),
	 "arcfour256": (EVP_RC4_KEY,'cipher_data'),
	 "aes128-cbc": (EVP_AES_KEY, 'cipher_data'), # aes*cbc == rijndael
	 "aes192-cbc": (EVP_AES_KEY, 'cipher_data'),
	 "aes256-cbc": (EVP_AES_KEY, 'cipher_data'),
	 "rijndael-cbc@lysator.liu.se": (ssh_rijndael_ctx, 'cipher_data'),
	 "aes128-ctr": (ssh_aes_ctr_ctx, 'app_data'),
	 "aes192-ctr": (ssh_aes_ctr_ctx, 'app_data'),
	 "aes256-ctr": (ssh_aes_ctr_ctx, 'app_data'),
	 "acss@openssh.org": None,
  }
  def loadMembers(self, mappings, maxDepth):
    if not LoadableMembers.loadMembers(self, mappings, maxDepth):
      return False
    #print 'app_data attr_obj_address=',getaddress(self.evp.app_data)
    # cast evp.app_data into a valid struct
    if self.cipher.contents.name.string in self.cipherContexts:
      struct,fieldname=self.cipherContexts[self.cipher.contents.name.string]
      if(struct is None):
        log.warning("Unsupported cipher %s"%(self.cipher.contents.name.string))
        return True
      attr=getattr(self.evp,fieldname)
      attr_obj_address=getaddress(attr)
      #print attr
      memoryMap = is_valid_address_value( attr_obj_address, mappings, struct)
      log.debug( "CAST %s into : %s "%(fieldname, struct) )
      st=struct.from_buffer_copy(memoryMap.readStruct(attr_obj_address, struct ) )
      # XXX CAST do not copy buffer when casting, sinon on perds des bytes
      attr.contents=(type(attr.contents)).from_buffer(st)

      log.debug('LOADED app_data evp.%s as %s from 0x%lx (%s) into 0x%lx'%(fieldname,struct, 
            attr_obj_address, is_valid_address_value(attr_obj_address,mappings,struct), getaddress(attr) ))
      log.debug('\t\t---------\n%s\t\t---------'%st.toString())
    else:
      log.warning("Unknown cipher %s, can't load a data struct for the EVP_CIPHER_CTX->app_data"%(self.cipher.contents.name.string))
    return True
  # MACRO
  def getEvpAppData(self):
    
    if self.cipher.contents.name.string in self.cipherContexts:
      struct,fieldname=self.cipherContexts[self.cipher.contents.name.string]
      if(struct is None):
        log.warning("Unsupported cipher %s"%(self.cipher.contents.name.string))
        log.warning("%s"%(self.cipher.contents.toString()))
        return None
      log.debug('CAST evp.%s Into %s'%(fieldname,struct))
      attr=getattr(self.evp,fieldname)
      st=struct.from_address(getaddress(attr))
      #print st.toString()
      return st
    return None

class Enc(OpenSSHStruct):
  ''' kex.h:84 '''
  _fields_ = [
  ("name",  CString ),  
  ("cipher", ctypes.POINTER(Cipher)),
  ("enabled",  ctypes.c_int), 
  ("key_len",  ctypes.c_uint), 
  ("block_size",  ctypes.c_uint), 
  ("key",  ctypes.POINTER(ctypes.c_ubyte)), #u_char ? -> ctypes.c_ubyte_p ?
  ("iv",  ctypes.POINTER(ctypes.c_ubyte))
  ]
  def loadMembers(self, mappings, maxDepth):
    if not LoadableMembers.loadMembers(self, mappings, maxDepth):
      return False
    # Load and memcopy key and iv
    log.debug('Memcopying a Key with %d bytes'%self.key_len)
    attr_obj_address=getaddress(self.key)
    memoryMap = is_valid_address_value( attr_obj_address, mappings)
    array=(ctypes.c_ubyte*self.key_len).from_buffer_copy(memoryMap.readArray(attr_obj_address, ctypes.c_ubyte, self.key_len))
    self.key.contents=ctypes.c_ubyte.from_buffer(array)
    
    log.debug('Memcopying a IV with %d bytes'%( self.block_size) )
    attr_obj_address=getaddress(self.iv)
    memoryMap = is_valid_address_value( attr_obj_address, mappings)
    array=(ctypes.c_ubyte*self.block_size).from_buffer_copy(memoryMap.readArray(attr_obj_address, ctypes.c_ubyte,self.block_size))
    self.iv.contents=ctypes.c_ubyte.from_buffer(array)
    
    log.debug('ENC KEY(%d bytes) and IV(%d bytes) acquired'%(self.key_len,self.block_size))
    return True
  def getKey(self):
    return pointer2bytes(self.key, self.key_len)
  def getIV(self):
    return pointer2bytes(self.iv, self.block_size) 

  def toPyObject(self):
    d=OpenSSHStruct.toPyObject(self)
    d.key = self.getKey()
    d.iv = self.getIV()
    return d
  
  def toString(self,prefix=''):
    s="%s # %s\n"%(prefix,repr(self) )
    for field,typ in self._fields_:
      attr=getattr(self,field)
      if field == 'key':
        #s+=prefix+'"%s": %s\n'%(field, self.getKey() )  
        s+=self._attrToString(self.getKey(),field,typ,prefix)
      elif field == 'iv':
        #s+=prefix+'"%s": %s\n'%(field, self.getIV() )  
        s+=self._attrToString(self.getIV(),field,typ,prefix)
      else:
        s+=self._attrToString(attr,field,typ,prefix)
    return s
  
class nh_ctx(OpenSSHStruct):
  ''' umac.c:323 '''
  _fields_ = [
  ("nh_key",  UINT8 *(L1_KEY_LEN + L1_KEY_SHIFT * (STREAMS - 1)) ), 
  ("data",  UINT8 * HASH_BUF_BYTES), 
  ("next_data_empty",  ctypes.c_int), 
  ("bytes_hashed",  ctypes.c_int), 
  ("state",  UINT64 * STREAMS)
  ]

class uhash_ctx(OpenSSHStruct):
  ''' umac.c:772 '''
  _fields_ = [
  ("hash",  nh_ctx), 
  ("poly_key_8",  UINT64 * STREAMS), 
  ("poly_accum",  UINT64 * STREAMS), 
  ("ip_keys",  UINT64 * STREAMS * 4), 
  ("ip_trans",  UINT32 * STREAMS), 
  ("msg_len",  UINT32)
  ]


class pdf_ctx(OpenSSHStruct):
  ''' umac:221 '''
  _fields_ = [
  ("cache",  UINT8 * AES_BLOCK_LEN), #UINT8 
  ("nonce",  UINT8 * AES_BLOCK_LEN), #UINT8 
  ("prf_key",  AES_KEY * 1) #typedef AES_KEY aes_int_key[1];
  ]

class umac_ctx(OpenSSHStruct):
  ''' umac:1179 '''
  _fields_ = [
  ("hash",  uhash_ctx), 
  ("pdf",  pdf_ctx), 
  ("free_ptr",  ctypes.c_void_p)
  ]

class Mac(OpenSSHStruct):
  ''' kex.h:90 '''
  _fields_ = [
  ("name",  CString ),  
  ("enabled",  ctypes.c_int), 
  ("mac_len",  ctypes.c_uint), 
  ("key",  ctypes.POINTER(ctypes.c_ubyte)), #u_char ? 
  ("key_len",  ctypes.c_uint), 
  ("type",  ctypes.c_int), 
  ("evp_md",  ctypes.POINTER(EVP_MD)),
  ("evp_ctx",  HMAC_CTX),
  ("umac_ctx",  ctypes.POINTER(umac_ctx)) 
  ]
  expectedValues={
    "type": [0,1,2], # unknown, SSH_EVP or SSH_UMAC
    "enabled": [0,1], # 
    "umac_ctx": IgnoreMember, # packet.c:1781
  }
  ''' if mac.type == SSH_EVP: (1) # mac.c:47
        evp_ctx is loaded/used when 
      else if mac.type == SSH_UMAC (2)
        umac_ctx is loaded/used

     We should conditionnally loadMembers on evp_ctx or umac_ctx, but, hey.. poc here...
  '''
  def loadMembers(self, mappings, maxDepth):
    if not LoadableMembers.loadMembers(self, mappings, maxDepth):
      return False
    # Load and memcopy key 
    log.debug('Memcopying a Key with %d bytes'%self.key_len)
    attr_obj_address=getaddress(self.key)
    memoryMap = is_valid_address_value( attr_obj_address, mappings)    
    array=(ctypes.c_ubyte*self.key_len).from_buffer_copy(memoryMap.readArray(attr_obj_address, ctypes.c_ubyte, self.key_len))
    self.key.contents=ctypes.c_ubyte.from_buffer(array)
    log.debug('unmac_ctx has been nulled and ignored. its not often used by any ssh impl. Not useful for us anyway.')
    log.debug('MAC KEY(%d bytes) acquired'%(self.key_len))
    return True
  def getKey(self):
    return pointer2bytes(self.key,self.key_len)

  def toPyObject(self):
    d=OpenSSHStruct.toPyObject(self)
    d.key = self.getKey()
    return d

  def toString(self,prefix=''):
    s="%s # %s\n"%(prefix,repr(self) )
    for field,typ in self._fields_:
      attr=getattr(self,field)
      if field == 'key':
        #s+=prefix+'"%s": %s\n'%(field, self.getKey() )  
        s+=self._attrToString(self.getKey(),field,typ,prefix)
      else:
        s+=self._attrToString(attr,field,typ,prefix)
    return s

class Comp(OpenSSHStruct):
  ''' kex.h:100 '''
  _fields_ = [
  ("type",  ctypes.c_int), 
  ("enabled",  ctypes.c_int), 
  ("name",  CString ) 
  ]

class Newkeys(OpenSSHStruct):
  ''' kex.h:110 '''
  _fields_ = [
  ("enc",  Enc), 
  ("mac",  Mac), 
  ("comp",  Comp)
  ]

class Buffer(OpenSSHStruct):
  ''' buffer.h:19 '''
  _fields_ = [
  ("buf", ctypes.POINTER(ctypes.c_ubyte) ), 
  ("alloc", ctypes.c_uint ), 
  ("offset", ctypes.c_uint ), 
  ("end", ctypes.c_uint)
  ]


class packet_state(OpenSSHStruct):
  ''' packet.c:90 '''
  _fields_ = [
  ("seqnr", UINT32 ), 
  ("packets", UINT32 ), 
  ("blocks", UINT64 ), 
  ("bytes", UINT64 )
  ]

class packet(OpenSSHStruct):
  pass
  

class TAILQ_HEAD_PACKET(OpenSSHStruct):
  ''' sys/queue.h:382 '''
  _fields_ = [
  ("tqh_first", ctypes.POINTER(packet) ), 
  ("tqh_last", ctypes.POINTER(ctypes.POINTER(packet)) )
  ]

class TAILQ_ENTRY_PACKET(OpenSSHStruct):
  ''' sys/queue.h:382 '''
  _fields_ = [
  ("tqe_next", ctypes.POINTER(packet) ), 
  ("tqe_prev", ctypes.POINTER(ctypes.POINTER(packet)) )
  ]

''' packet.c:90 '''
packet._fields_ = [
  ("next", TAILQ_ENTRY_PACKET), 
  ("type", ctypes.c_ubyte ), #u_char
  ("payload", Buffer )
  ] 

packet.expectedValues={
    "next": IgnoreMember, 
}
'''
TAILQ_HEAD_PACKET.expectedValues={
    "next": IgnoreMember, 
}
.expectedValues={
    "next": IgnoreMember, 
}
'''
class session_state(OpenSSHStruct):
  ''' openssh/packet.c:103 '''
  _fields_ = [
  ("connection_in", ctypes.c_int ), 
  ("connection_out", ctypes.c_int ), 
  ("remote_protocol_flags", ctypes.c_uint ), 
  ("receive_context", CipherContext ), # used to cipher_crypt/receive
  ("send_context", CipherContext ),    # used to cipher_crypt/send
  ("input", Buffer ), 
  ("output", Buffer ), 
  ("outgoing_packet", Buffer ), 
  ("incoming_packet", Buffer ), 
  ("compression_buffer", Buffer ), 
  ("compression_buffer_ready", ctypes.c_int ), 
  ("packet_compression", ctypes.c_int ), 
  ("max_packet_size", ctypes.c_uint ), 
  ("initialized", ctypes.c_int ), 
  ("interactive_mode", ctypes.c_int ), 
  ("server_side", ctypes.c_int ), 
  ("after_authentication", ctypes.c_int ), 
  ("keep_alive_timeouts", ctypes.c_int ), 
  ("packet_timeout_ms", ctypes.c_int ), 
  ("newkeys", ctypes.POINTER(Newkeys)*MODE_MAX ), #Newkeys *newkeys[MODE_MAX]; XXX
  ("p_read", packet_state ), 
  ("p_send", packet_state ), 
  ("max_blocks_in", UINT64 ), 
  ("max_blocks_out", UINT64 ), 
  ("rekey_limit", UINT32 ), 
  ("ssh1_key", ctypes.c_ubyte * SSH_SESSION_KEY_LENGTH ), #	u_char ssh1_key[SSH_SESSION_KEY_LENGTH];
  ("ssh1_keylen", ctypes.c_uint ), 
  ("extra_pad", ctypes.c_ubyte ), 
  ("packet_discard", ctypes.c_uint ), 
  ("packet_discard_mac", ctypes.POINTER(Mac) ), 
  ("packlen", ctypes.c_uint ), 
  ("rekeying", ctypes.c_int ), 
  ("set_interactive_called", ctypes.c_int ), 
  ("set_maxsize_called", ctypes.c_int ), 
  ("outgoing", TAILQ_HEAD_PACKET ) 
  ]
  expectedValues={
    "connection_in": RangeValue(-1,1024), # FD number
    "connection_out": RangeValue(-1,1024), # FD number
    "max_packet_size": RangeValue(4 * 1024,1024 * 1024), # packet.c:1781
    "packlen": RangeValue(0,1024 * 1024), # is 0 or < max_pack_size
    "initialized": [1], # mostly should be 1
    "interactive_mode": [0,1], 
    "server_side": [0,1] , 
    "after_authentication": [0,1], # mostly should be 1
    #"keep_alive_timeouts", [0], ## ?? 
    #"packet_timeout_ms", [], 
    #"max_blocks_in", UINT64 , #packet.c:794 1<<enc.block_size*2   ou si block_size < 16 (1<<30)/enc.block_size
    #"max_blocks_out", UINT64 , # ou si rekey_limit , max_blocks = MIN(max_block, rekey_limit/block_size )
    #"max_blocks_in": NotNull, #mmh # 0L values in sshd server
    #"max_blocks_out": NotNull # ) value in sshd server
  }
  def toPyObject(self):
    d=OpenSSHStruct.toPyObject(self)
    # populate AppData.
    d.receive_context.evp.app_data = self.receive_context.getEvpAppData().toPyObject()
    d.send_context.evp.app_data = self.send_context.getEvpAppData().toPyObject()
    ## TODOm find a better way to pass a void_p for that cipher data
    d.receive_context.evp.cipher_data = self.receive_context.getEvpAppData().toPyObject()
    d.send_context.evp.cipher_data = self.send_context.getEvpAppData().toPyObject()
    return d


def printSizeof():
  print 'Cipher:',ctypes.sizeof(Cipher)
  print 'CipherContext:',ctypes.sizeof(CipherContext)
  print 'Enc:',ctypes.sizeof(Enc)
  print 'nh_ctx:',ctypes.sizeof(nh_ctx)
  print 'uhash_ctx:',ctypes.sizeof(uhash_ctx)
  print 'pdf_ctx:',ctypes.sizeof(pdf_ctx)
  print 'umac_ctx:',ctypes.sizeof(umac_ctx)
  print 'Mac:',ctypes.sizeof(Mac)
  print 'Comp:',ctypes.sizeof(Comp)
  print 'Newkeys:',ctypes.sizeof(Newkeys)
  print 'Buffer:',ctypes.sizeof(Buffer)
  print 'packet:',ctypes.sizeof(packet)
  print 'packet_state:',ctypes.sizeof(packet_state)
  print 'TAILQ_HEAD_PACKET:',ctypes.sizeof(TAILQ_HEAD_PACKET)
  print 'TAILQ_ENTRY_PACKET:',ctypes.sizeof(TAILQ_ENTRY_PACKET)
  print 'session_state:',ctypes.sizeof(session_state)
  print 'UINT32:',ctypes.sizeof(UINT32)
  print 'UINT64:',ctypes.sizeof(UINT64)
  print 'UINT8:',ctypes.sizeof(UINT8)
  print 'AES_BLOCK_LEN:',AES_BLOCK_LEN
  print 'HASH_BUF_BYTES:',HASH_BUF_BYTES
  print 'UMAC_OUTPUT_LEN:',UMAC_OUTPUT_LEN
  print 'SSH_SESSION_KEY_LENGTH:',SSH_SESSION_KEY_LENGTH
  print 'L1_KEY_LEN:',L1_KEY_SHIFT
  print 'L1_KEY_SHIFT:',L1_KEY_SHIFT
  print 'MODE_MAX:',MODE_MAX
  print 'STREAMS:',STREAMS


import inspect,sys
''' Load all openSSH classes and used OpenSSL classes to local classRef '''
OpenSSHStruct.classRef=dict([ (ctypes.POINTER( klass), klass) for (name,klass) in inspect.getmembers(sys.modules[__name__], inspect.isclass) if klass.__module__ == __name__ or klass.__module__ == 'ctypes_openssl'])

''' Load all model classes and create a similar non-ctypes Python class  
  thoses will be used to translate non pickable ctypes into POPOs.
'''
## model.createPOPO(__name__)
for klass,typ in inspect.getmembers(sys.modules[__name__], inspect.isclass):
  if typ.__module__ == __name__:
    setattr(sys.modules[__name__], '%s_py'%(klass), type('%s_py'%(klass),(model.pyObj,),{}) )
    #print sys.modules[__name__], '%s_py'%(klass)

if __name__ == '__main__':
  printSizeof()






#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2011 Loic Jaquemet loic.jaquemet+python@gmail.com
#

__author__ = "Loic Jaquemet loic.jaquemet+python@gmail.com"

import ctypes
import logging

from ptrace.debugger.memory_mapping import readProcessMappings

from haystack.model import is_valid_address,is_valid_address_value,getaddress,array2bytes,bytes2array
from haystack.model import LoadableMembers,RangeValue,NotNull,CString,EVP_CIPHER_CTX_APP_DATA_PTR
from haystack import model

log=logging.getLogger('openssl.model')

''' hmac.h:69 '''
HMAC_MAX_MD_CBLOCK=128
''' evp.h:91 '''
EVP_MAX_BLOCK_LENGTH=32
EVP_MAX_IV_LENGTH=16
AES_MAXNR=14 # aes.h:66
RIJNDAEL_MAXNR=14

# we have to alloc a big chunk
BN_ULONG=ctypes.c_ulong

class OpenSSLStruct(LoadableMembers):
  ''' defines classRef '''
  pass




class rijndael_ctx(OpenSSLStruct):
  _fields_ = [
  ('decrypt',ctypes.c_int),
  ('Nr',ctypes.c_int),
  ('ek',  (ctypes.c_uint32*(4*RIJNDAEL_MAXNR + 1))  ),
  ('dk',  (ctypes.c_uint32*(4*RIJNDAEL_MAXNR + 1))  ),
  ]  

class RC4_KEY(OpenSSLStruct):
  _fields_ = [
  ('x',ctypes.c_uint), #RC4_INT opensslconf.h:61
  ('x',ctypes.c_uint),
  ('data',ctypes.c_uint*256)
  ]

class EVP_RC4_KEY(OpenSSLStruct):
  _fields_ = [
  ('ks',RC4_KEY)
  ]

class AES_KEY(OpenSSLStruct):
  ''' aes.h:78 '''
  _fields_ = [
  ("rd_key",  ctypes.c_ulong * (4* (AES_MAXNR+1))), 
  ("rounds",  ctypes.c_int)
  ] 
  def getKey(self):
    #return array2bytes(self.rd_key)
    return ','.join(["0x%lx"%key for key in self.rd_key])
  def getRounds(self):
    return self.rounds
  def fromPyObj(self,pyobj):
    #copy rd_key
    self.rd_key=bytes2array(pyobj.rd_key,ctypes.c_ulong)
    #copy rounds
    self.rounds=pyobj.rounds
    return self

#test
class EVP_AES_KEY(OpenSSLStruct):
  _fields_ = [
  ('ks', AES_KEY),
	]
  def fromPyObj(self,pyobj):
    self.ks = AES_KEY().fromPyObj(pyobj.ks)
    return self

#ok
class BIGNUM(OpenSSLStruct):
  _fields_ = [
  ("d",ctypes.POINTER(BN_ULONG) ),
  ('top',ctypes.c_int),
  ('dmax',ctypes.c_int),
  ('neg',ctypes.c_int),
  ('flags',ctypes.c_int)
  ]
  expectedValues={
    "neg": [0,1]
  }
  def loadMembers(self, mappings, maxDepth):
    ''' 
    #self._d = process.readArray(attr_obj_address, ctypes.c_ulong, self.top) 
    ## or    
    #ulong_array= (ctypes.c_ulong * self.top)    
    '''
    if not self.isValid(mappings):
      log.debug('BigNUm tries to load members when its not validated')
      return False
    if True:
      # Load and memcopy d / BN_ULONG *
      attr_obj_address=getaddress(self.d)
      memoryMap = is_valid_address_value( attr_obj_address, mappings)
      #print memoryMap, type(memoryMap)
      contents=(BN_ULONG*self.top).from_buffer_copy(memoryMap.readArray(attr_obj_address, BN_ULONG, self.top))
      log.debug('contents acquired %d'%ctypes.sizeof(contents))
      self.d.contents=BN_ULONG.from_address(ctypes.addressof(contents))
      # TODO ctypes.from_address(address)
      #self.d=ctypes.cast(ctypes.pointer(contents), ctypes.POINTER(BN_ULONG) ) 
      self.d=ctypes.cast(contents, ctypes.POINTER(BN_ULONG) ) 
      return True
    else:
      ### we can't cast , so we need to copy ulong by ulong
      #if (ctypes.sizeof(ctypes.c_ulong)*self.top) >= ctypes.sizeof(self.d.contents):
      #  log.error('self.top %d is too big for us puny humans'%self.top)
      #  return False
      # memory leak surely, but we can't handle 
      top=self.top
      contents=BN_ULONG(0)
      self.d.contents=contents
      print '7 -> ',self
      print "self: 0x%lx"%ctypes.addressof(self)
      log.debug("contents : 0x%lx-0x%lx"%( ctypes.addressof(contents), ctypes.addressof(contents)+ ctypes.sizeof(contents) ))
      log.debug("contents : 0x%lx-0x%lx"%( ctypes.addressof(contents), ctypes.addressof(contents)+ ctypes.sizeof(contents) ))
      log.debug("d.contents : 0x%lx-0x%lx"%( ctypes.addressof(self.d.contents), ctypes.addressof(self.d.contents)+ ctypes.sizeof(self.d.contents) ))
      log.debug("contents : 0x%lx-0x%lx"%( ctypes.addressof(contents), ctypes.addressof(contents)+ ctypes.sizeof(contents) ))
      #log.debug('new memspace : 0x%lx with size %d '%(ctypes.addressof(self.d.contents), ctypes.sizeof(self.d.contents) )   )
      log.debug("contents : 0x%lx-0x%lx"%( getaddress(contents), getaddress(contents)+ ctypes.sizeof(contents) ))
      log.debug("d.contents : 0x%lx-0x%lx"%( getaddress(self.d.contents), getaddress(self.d.contents)+ ctypes.sizeof(self.d.contents) ))
      print '8 -> ',self
      log.debug("contents : 0x%lx-0x%lx"%( getaddress(contents), getaddress(contents)+ ctypes.sizeof(contents) ))
      log.debug("d.contents : 0x%lx-0x%lx"%( getaddress(self.d.contents), getaddress(self.d.contents)+ ctypes.sizeof(self.d.contents) ))
      print top
      for i in range(0,top):
        self.d.contents[i]=contents[i]
      log.debug('contents copied %d'%(i+1))
      log.debug('%s loaded at 0x%lx'%(self.__class__.__name__,ctypes.addressof(self)))    
      return True
  
  def isValid(self,mappings):
    if ( self.dmax < 0 or self.top < 0 or self.dmax < self.top ):
      return False
    return LoadableMembers.isValid(self,mappings)
  
  def __str__(self):
    d= getaddress(self.d)
    return ("BN { d=0x%lx, top=%d, dmax=%d, neg=%d, flags=%d }"%
                (d, self.top, self.dmax, self.neg, self.flags) )
#ok
class STACK(OpenSSLStruct):
  _fields_ = [
  ("num",ctypes.c_int),
  ("data",ctypes.POINTER(ctypes.c_ubyte)), 
  ("sorted",ctypes.c_int),
  ("num_alloc",ctypes.c_int),
  ("comp",ctypes.POINTER(ctypes.c_int) ) ]

#ok
class CRYPTO_EX_DATA(OpenSSLStruct):
  _fields_ = [
  ("sk",ctypes.POINTER(STACK) ),
  ("dummy",ctypes.c_int)]
  def loadMembers(self, mappings, maxDepth):
    ''' erase self.sk'''
    #self.sk=ctypes.POINTER(STACK)()
    return LoadableMembers.loadMembers(self, mappings, maxDepth)
  def isValid(self,mappings):
    ''' erase self.sk'''
    # TODO why ?
    #self.sk=ctypes.POINTER(STACK)()
    return LoadableMembers.isValid(self,mappings)
  
#ok
class BN_MONT_CTX(OpenSSLStruct):
  _fields_ = [
  ("ri",ctypes.c_int),
  ("RR",BIGNUM),
  ("N",BIGNUM),
  ("Ni",BIGNUM),
  ("n0",ctypes.c_ulong),
  ("flags",ctypes.c_int)]

class EVP_PKEY(OpenSSLStruct):
	_fields_ = [
  ('type',ctypes.c_int),
  ('save_type',ctypes.c_int),
  ('references',ctypes.c_int),
  ('pkey',ctypes.c_void_p), ## union of struct really
  ('save_parameters',ctypes.c_int),
  ('attributes',ctypes.c_void_p) ## 	STACK_OF(X509_ATTRIBUTE) *attributes; /* [ 0 ] */
  ]


class ENGINE_CMD_DEFN(OpenSSLStruct):
	_fields_ = [
  ('cmd_num',ctypes.c_uint),
  ('cmd_name', CString),
  ('cmd_desc', CString),
  ('cmd_flags',ctypes.c_uint)
  ]  

class ENGINE(OpenSSLStruct):
  pass
ENGINE._fields_ = [
  ('id', CString),
  ('name', CString),
  ('rsa_meth',ctypes.POINTER(ctypes.c_int) ),
  ('dsa_meth',ctypes.POINTER(ctypes.c_int) ),
  ('dh_meth',ctypes.POINTER(ctypes.c_int) ),
  ('ecdh_meth',ctypes.POINTER(ctypes.c_int) ),
  ('ecdsa_meth',ctypes.POINTER(ctypes.c_int) ),
  ('rand_meth',ctypes.POINTER(ctypes.c_int) ),
  ('store_meth',ctypes.POINTER(ctypes.c_int) ),
  ('ciphers',ctypes.POINTER(ctypes.c_int) ), ## fn typedef int (*ENGINE_CIPHERS_PTR)(ENGINE *, const EVP_CIPHER **, const int **, int);
  ('digest',ctypes.POINTER(ctypes.c_int) ),  ## fn typedef int (*ENGINE_DIGESTS_PTR)(ENGINE *, const EVP_MD **, const int **, int);
  ('destroy',ctypes.POINTER(ctypes.c_int) ), ## fn typedef int (*ENGINE_GEN_INT_FUNC_PTR)(ENGINE *);
  ('init',ctypes.POINTER(ctypes.c_int) ),    ## fn typedef int (*ENGINE_GEN_INT_FUNC_PTR)(ENGINE *);
  ('finish',ctypes.POINTER(ctypes.c_int) ),  ## fn typedef int (*ENGINE_GEN_INT_FUNC_PTR)(ENGINE *);
  ('ctrl',ctypes.POINTER(ctypes.c_int) ),    ## fn typedef int (*ENGINE_CTRL_FUNC_PTR)(ENGINE *, int, long, void *, void (*f)(void));
  ('load_privkey',ctypes.POINTER(EVP_PKEY) ), ## fn EVP_PKEY *
  ('load_pubkey',ctypes.POINTER(EVP_PKEY) ),  ## fn EVP_PKEY *
  ('load_ssl_client_cert',ctypes.POINTER(ctypes.c_int) ), ## fn typedef int (*ENGINE_SSL_CLIENT_CERT_PTR)(ENGINE *, SSL *ssl,
  ('cmd_defns',ctypes.POINTER(ENGINE_CMD_DEFN) ), ##
  ('flags',ctypes.c_int),
  ('struct_ref',ctypes.c_int),
  ('funct_ref',ctypes.c_int),
  ('ex_data',CRYPTO_EX_DATA),
  ('prev',ctypes.POINTER(ENGINE) ),
  ('nex',ctypes.POINTER(ENGINE) )
  ]


#KO
class RSA(OpenSSLStruct):
  ''' rsa/rsa.h '''
  loaded=False
  _fields_ = [
  ("pad",  ctypes.c_int), 
  ("version",  ctypes.c_long),
  ("meth",ctypes.POINTER(BIGNUM)),#const RSA_METHOD *meth;
  ("engine",ctypes.POINTER(ENGINE)),#ENGINE *engine;
  ('n', ctypes.POINTER(BIGNUM) ), ## still in ssh memap
  ('e', ctypes.POINTER(BIGNUM) ), ## still in ssh memap
  ('d', ctypes.POINTER(BIGNUM) ), ## still in ssh memap
  ('p', ctypes.POINTER(BIGNUM) ), ## still in ssh memap
  ('q', ctypes.POINTER(BIGNUM) ), ## still in ssh memap
  ('dmp1', ctypes.POINTER(BIGNUM) ),
  ('dmq1', ctypes.POINTER(BIGNUM) ),
  ('iqmp', ctypes.POINTER(BIGNUM) ),
  ("ex_data", CRYPTO_EX_DATA ),
  ("references", ctypes.c_int),
  ("flags", ctypes.c_int),
  ("_method_mod_n", ctypes.POINTER(BN_MONT_CTX) ),
  ("_method_mod_p", ctypes.POINTER(BN_MONT_CTX) ),
  ("_method_mod_q", ctypes.POINTER(BN_MONT_CTX) ),
  ("bignum_data",ctypes.POINTER(ctypes.c_ubyte)), ## moue c_char_p ou POINTER(c_char) ?
  ("blinding",ctypes.POINTER(BIGNUM)),#BN_BLINDING *blinding;
  ("mt_blinding",ctypes.POINTER(BIGNUM))#BN_BLINDING *mt_blinding;
  ]
  expectedValues={
    "pad": [0], 
    "version": [0], 
    "references": RangeValue(0,0xfff),
    "n": [NotNull],
    "e": [NotNull],
    "d": [NotNull],
    "p": [NotNull],
    "q": [NotNull],
    "dmp1": [NotNull],
    "dmq1": [NotNull],
    "iqmp": [NotNull]
  }
  def printValid(self,mappings):
    log.debug( '----------------------- LOADED: %s'%self.loaded)
    log.debug('pad: %d version %d ref %d'%(self.pad,self.version,self.references) )
    log.debug(is_valid_address( self.n, mappings)    )
    log.debug(is_valid_address( self.e, mappings)    )
    log.debug(is_valid_address( self.d, mappings)    )
    log.debug(is_valid_address( self.p, mappings)    )
    log.debug(is_valid_address( self.q, mappings)    )
    log.debug(is_valid_address( self.dmp1, mappings) ) 
    log.debug(is_valid_address( self.dmq1, mappings) )
    log.debug(is_valid_address( self.iqmp, mappings) )
    return
  def loadMembers(self, mappings, maxDepth):
    # XXXX clean other structs
    self.meth = None
    #self._method_mod_n = ctypes.POINTER(BN_MONT_CTX)()
    #self._method_mod_p = ctypes.POINTER(BN_MONT_CTX)()
    #self._method_mod_q = ctypes.POINTER(BN_MONT_CTX)()
    self.bignum_data = None
    self.blinding = None
    self.mt_blinding = None

    if not LoadableMembers.loadMembers(self, mappings, maxDepth):
      log.debug('RSA not loaded')
      return False
    return True
    
#KO
class DSA(OpenSSLStruct):
  _fields_ = [
  ("pad",  ctypes.c_int), 
  ("version",  ctypes.c_long),
  ("write_params",ctypes.c_int),
  ('p', ctypes.POINTER(BIGNUM) ),
  ('q', ctypes.POINTER(BIGNUM) ),
  ('g', ctypes.POINTER(BIGNUM) ),
  ('pub_key', ctypes.POINTER(BIGNUM) ),
  ('priv_key', ctypes.POINTER(BIGNUM) ),
  ('kinv', ctypes.POINTER(BIGNUM) ),
  ('r', ctypes.POINTER(BIGNUM) ),
  ("flags", ctypes.c_int),
  ("_method_mod_p", ctypes.POINTER(BN_MONT_CTX) ),
  ("references", ctypes.c_int),
  ("ex_data", CRYPTO_EX_DATA ),
  ("meth",ctypes.POINTER(ctypes.c_int)),#  const DSA_METHOD *meth;
  ("engine",ctypes.POINTER(ENGINE))
  ]
  expectedValues={
    "pad": [0], 
    "version": [0], 
    "references": RangeValue(0,0xfff),
    "p": [NotNull],
    "q": [NotNull],
    "g": [NotNull],
    "pub_key": [NotNull],
    "priv_key": [NotNull]
  }
  def printValid(self,mappings):
    log.debug( '----------------------- \npad: %d version %d ref %d'%(self.pad,self.version,self.write_params) )
    log.debug(is_valid_address( self.p, mappings)    )
    log.debug(is_valid_address( self.q, mappings)    )
    log.debug(is_valid_address( self.g, mappings)    )
    log.debug(is_valid_address( self.pub_key, mappings)    )
    log.debug(is_valid_address( self.priv_key, mappings)    )
    return
  def internalCheck(self):
    '''  pub_key = g^privKey mod p '''
    return

  def loadMembers(self, mappings, maxDepth):
    # clean other structs
    # r and kinv can be null
    self.meth = None
    self._method_mod_p = None
    #self.engine = None
    
    if not LoadableMembers.loadMembers(self, mappings, maxDepth):
      log.debug('DSA not loaded')
      return False

    return True

#ok
class EVP_CIPHER(OpenSSLStruct):
  ''' evp.h:332 '''	
  _fields_ = [
  ("nid",  ctypes.c_int), 
  ("block_size",  ctypes.c_int), 
  ("key_len",  ctypes.c_int), 
  ("iv_len",  ctypes.c_int), 
  ("flags",  ctypes.c_ulong), 
  ("init",  ctypes.POINTER(ctypes.c_int)), # function () 
  ("do_cipher",  ctypes.POINTER(ctypes.c_int)), # function () ## crypt func.
  ("cleanup",  ctypes.POINTER(ctypes.c_int)), # function () 
  ("ctx_size",  ctypes.c_int), 
  ("set_asn1_parameters",  ctypes.POINTER(ctypes.c_int)), # function () 
  ("get_asn1_parameters",  ctypes.POINTER(ctypes.c_int)), # function () 
  ("ctrl",  ctypes.POINTER(ctypes.c_int)), # function () 
  ("app_data",  ctypes.POINTER(ctypes.c_ubyte)) 
  ]
  expectedValues={
    "key_len": RangeValue(1,0xff), # key_len *8 bits ..2040 bits for a key is enought ? 
                                   # Default value for variable length ciphers 
    "iv_len": RangeValue(1,0xff), #  
    "init": [NotNull], 
    "do_cipher": [NotNull], 
    #"cleanup": [NotNull], # aes-cbc ?
    "ctx_size": RangeValue(0,0xffff), #  app_data struct should not be too big
  }

#mok
class EVP_CIPHER_CTX(OpenSSLStruct):
  ''' evp.h:332 '''	
  _fields_ = [
  ("cipher",  ctypes.POINTER(EVP_CIPHER)), 
  ("engine",  ctypes.POINTER(ctypes.c_int)), ## TODO ENGINE*
  ("encrypt",  ctypes.c_int), 
  ("buf_len",  ctypes.c_int), 
  ("oiv",  ctypes.c_ubyte*EVP_MAX_IV_LENGTH),## unsigned char  oiv[EVP_MAX_IV_LENGTH];
  ("iv",  ctypes.c_ubyte*EVP_MAX_IV_LENGTH), ##unsigned char  iv[EVP_MAX_IV_LENGTH];
  ("buf",  ctypes.c_ubyte*EVP_MAX_BLOCK_LENGTH), ##unsigned char buf[EVP_MAX_BLOCK_LENGTH];
  ("num",  ctypes.c_int), 
  ("app_data",  EVP_CIPHER_CTX_APP_DATA_PTR), # utilise par ssh_aes/rijndael
  ("key_len",  ctypes.c_int), 
  ("flags",  ctypes.c_ulong), 
  ("cipher_data",  EVP_CIPHER_CTX_APP_DATA_PTR), ## utilise par rc4 ?
  ("final_used",  ctypes.c_int), 
  ("block_mask",  ctypes.c_int), 
  ("final",  ctypes.c_ubyte*EVP_MAX_BLOCK_LENGTH) ###unsigned char final[EVP_MAX_BLOCK_LENGTH]
  ]
  expectedValues={
    "cipher": [NotNull], 
    "encrypt": [0,1], 
    "buf_len": RangeValue(0,EVP_MAX_BLOCK_LENGTH), ## number we have left, so must be less than buffer_size
    #"engine": , # can be null
    #"app_data": , # can be null if cipher_data is not
    #"cipher_data": , # can be null if app_data is not
    "key_len": RangeValue(1,0xff), # key_len *8 bits ..2040 bits for a key is enought ? 
  }
  def getOIV(self):
    return array2bytes(self.oiv)
  def getIV(self):
    return array2bytes(self.iv)
  def getAppData(self,structType):
    log.debug('CAST app_data into %s'%(structType))
    return structType.from_address(getaddress(self.app_data))

#mok
class EVP_MD(OpenSSLStruct):
  ''' struct env_md_st evp.h:227 '''
  _fields_ = [
  ("type",  ctypes.c_int), 
  ("pkey_type",  ctypes.c_int), 
  ("md_size",  ctypes.c_int), 
  ("flags",  ctypes.c_ulong), 
  ("init",  ctypes.POINTER(ctypes.c_int)), # function () 
  ("update",  ctypes.POINTER(ctypes.c_int)), # function () 
  ("final",  ctypes.POINTER(ctypes.c_int)), # function () 
  ("copy",  ctypes.POINTER(ctypes.c_int)), # function () 
  ("cleanup",  ctypes.POINTER(ctypes.c_int)), # function () 
  ("sign",  ctypes.POINTER(ctypes.c_int)), # function () 
  ("verify",  ctypes.POINTER(ctypes.c_int)), # function () 
  ("required_pkey_type",  ctypes.c_int*5), #required_pkey_type[5]
  ("block_size",  ctypes.c_int), 
  ("ctx_size",  ctypes.c_int)
  ]

class EVP_MD_CTX(OpenSSLStruct):
  ''' evp.h:304 '''
  _fields_ = [
  ("digest",  ctypes.POINTER(EVP_MD)),
  ("engine",  ctypes.POINTER(ENGINE) ), #
  ("flags",  ctypes.c_ulong),
  ("md_data",  ctypes.POINTER(ctypes.c_ubyte))
  ]

class HMAC_CTX(OpenSSLStruct):
  ''' hmac.h:75 '''
  _fields_ = [
  ("md",  ctypes.POINTER(EVP_MD)), 
  ("md_ctx",  EVP_MD_CTX), 
  ("i_ctx",  EVP_MD_CTX), 
  ("o_ctx",  EVP_MD_CTX), 
  ("key_length",  ctypes.c_uint), 
  ("key",  ctypes.c_char * HMAC_MAX_MD_CBLOCK)
  ] 


def printSizeof():
  print 'BIGNUM:',ctypes.sizeof(BIGNUM)
  print 'STACK:',ctypes.sizeof(STACK)
  print 'CRYPTO_EX_DATA:',ctypes.sizeof(CRYPTO_EX_DATA)
  print 'BN_MONT_CTX:',ctypes.sizeof(BN_MONT_CTX)
  print 'EVP_PKEY:',ctypes.sizeof(EVP_PKEY)
  print 'ENGINE_CMD_DEFN:',ctypes.sizeof(ENGINE_CMD_DEFN)
  print 'ENGINE:', ctypes.sizeof(ENGINE)
  print 'RSA:',ctypes.sizeof(RSA)
  print 'DSA:',ctypes.sizeof(DSA)
  print 'EVP_CIPHER:',ctypes.sizeof(EVP_CIPHER)
  print 'EVP_CIPHER_CTX:',ctypes.sizeof(EVP_CIPHER_CTX)
  print 'EVP_MD:',ctypes.sizeof(EVP_MD)
  print 'EVP_MD_CTX:',ctypes.sizeof(EVP_MD_CTX)
  print 'HMAC_CTX:',ctypes.sizeof(HMAC_CTX)
  print 'AES_KEY:',ctypes.sizeof(AES_KEY) #AES_KEY: 232 au lieu de 244 .. chiotote..
  print 'HMAC_MAX_MD_CBLOCK:',HMAC_MAX_MD_CBLOCK
  print 'EVP_MAX_BLOCK_LENGTH:',EVP_MAX_BLOCK_LENGTH
  print 'EVP_MAX_IV_LENGTH:',EVP_MAX_IV_LENGTH
  print 'AES_MAXNR:',AES_MAXNR

import inspect,sys
''' Load all openSSL classes to local classRef '''
OpenSSLStruct.classRef=dict([ (ctypes.POINTER( klass), klass) for (name,klass) in inspect.getmembers(sys.modules[__name__], inspect.isclass) if klass.__module__ == __name__])

''' Load all model classes and create a similar non-ctypes Python class  
  thoses will be used to translate non pickable ctypes into POPOs.
'''
for klass,typ in inspect.getmembers(sys.modules[__name__], inspect.isclass):
  if typ.__module__ == __name__:
    setattr(sys.modules[__name__], '%s_py'%(klass), type('%s_py'%(klass),(model.pyObj,),{}) )

if __name__ == '__main__':
  printSizeof()




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2011 Loic Jaquemet loic.jaquemet+python@gmail.com
#

__author__ = "Loic Jaquemet loic.jaquemet+python@gmail.com"

import argparse, os, logging, sys, time, pickle, struct

import ctypes
import ctypes_openssh, ctypes_openssl
import output, socket_scapy

from engine import CIPHERS
import haystack 

#our impl
from paramiko_packet import Packetizer

# todo : replace by one empty shell of ours
from paramiko.transport import Transport

from output import FileWriter

log=logging.getLogger('sslsnoop.openssh')



CLIENT_STRUCTS=[ctypes_openssh.session_state]
SERVER_STRUCTS=[ctypes_openssh.session_state]
AGENT_STRUCTS=[ctypes_openssl.RSA, ctypes_openssl.DSA]



class SessionStateFileWriter(FileWriter):
  def __init__(self,pid,folder='outputs'):
    FileWriter.__init__(self,'session_state',pid,folder)
  def writeToFile(self,instance):
    prefix=self.prefix
    filename=self.get_valid_filename()
    f=open(filename,"w")
    pickle.dump(instance,f)
    f.close()
    log.info ("[X] SSH session_state saved to file %s"%filename)
    return True




class DumbCipherContext:
  pass

class SessionCiphers():
  def __init__(self,session_state):
    self.session_state=session_state
    # two ciphers    
    self.receiveCtx=DumbCipherContext()
    self.sendCtx=DumbCipherContext()
    # read cipher MODE_IN == 0
    MODE=0
    for ctx in [self.receiveCtx, self.sendCtx ]:
      ctx.enc =  self.session_state.newkeys[MODE].enc
      ctx.mac =  self.session_state.newkeys[MODE].mac
      ctx.comp = self.session_state.newkeys[MODE].comp
      if MODE == 0:
        ctx.sshCtx = session_state.receive_context
      else:
        ctx.sshCtx = session_state.send_context
      ctx.sshCipher = ctx.sshCtx.cipher
      ctx.evpCtx    = ctx.sshCtx.evp
      ctx.evpCipher = ctx.sshCtx.evp.cipher
      # useful stuff
      ctx.name = ctx.sshCipher.name
      ctx.app_data  = ctx.evpCtx.app_data  #TODO delete
      ctx.cipher_data  = ctx.evpCtx.cipher_data #TODO delete
      # original key and IV are ctx.getKey() and ctx.getIV()
      # stateful AES_key key is at ctx.app_data.aes_ctx #&c->aes_ctx
      # stateful ctr counter is at ctx.app_data.aes_ctr
      ctx.key_len  = ctx.evpCipher.key_len
      ctx.block_size  = ctx.evpCipher.block_size
      MODE+=1
    # 
    return

  def getCiphers(self):
    return self.receiveCtx, self.sendCtx
  
  def __str__(self):
    return "<SessionCiphers RECEIVE: '%s/%s' SEND: '%s/%s' >"%(self.receiveCtx.name,self.receiveCtx.mac.name,
                                                  self.sendCtx.name,self.sendCtx.mac.name ) 



class OpenSSHKeysFinder():
  ''' wrapper around a fork/exec to abouchet StructFinder '''

  def __init__(self, pid, fullScan=False):
    self.pid = pid
    self.fullScan = fullScan
    return
  
  def find_struct(self, typ, maxNum):
    raise NotImplementedError()
    return

  def loadAt(self,offset, typ):
    raise NotImplementedError()
    return
  
  def findActiveSession(self, maxNum=1):
    ''' '''
    outs=haystack.findStruct(self.pid, ctypes_openssh.session_state)
    if outs is None:
      log.error("The session_state has not been found. maybe it's not OpenSSH ?")
      return None,None
    elif len(outs) > 1:
      log.warning("Mmmh, we found multiple session_state(%d). That is odd. I'll try with the first one."%(len(outs)))
    #
    session_state,addr=outs[0]
    return session_state,addr

  def refreshActiveSession(self, offset):
    ''' '''
    instance,validated=haystack.refreshStruct(self.pid, ctypes_openssh.session_state, offset)
    if not validated:
      log.error("The session_state has not been re-validated. You should look for it again.")
      return None,None
    return instance,offset
    
  def findActiveKeys(self, maxNum=1, offset=None):
    ''' '''
    if offset is None:
      session_state,addr=self.findActiveSession(maxNum)
    else:
      session_state,addr=self.refreshActiveSession(offset)
    if session_state is None:
      return None,None # raise Exception ... 
    ciphers=SessionCiphers(session_state)
    log.info('Active state ciphers : %s at 0x%lx'%(ciphers,addr))
    #log.debug(ciphers.receiveCtx.app_data.__dict__)
    return ciphers,addr

  def save(self, instance):
    ssfw=SessionStateFileWriter(self.process.pid)
    ssfw.writeToFile(instance)
    return






class OpenSSHLiveDecryptatator(OpenSSHKeysFinder):
  ''' Decrypt SSH traffic in live '''
  def __init__(self, pid, sessionStateAddr=None, scapySocketThread = None, serverMode=None, fullScan=False, maxNum=1):
    OpenSSHKeysFinder. __init__(self, pid, fullScan=fullScan)
    self.soscapy=scapySocketThread
    self.serverMode=serverMode
    self.maxNum = maxNum
    self.session_state_addr=sessionStateAddr
    self.inbound=dict()
    self.outbound=dict()
    return
  
  def run(self):
    ''' launch sniffer and decrypter threads '''
    if self.soscapy is None:
      self.soscapy=launchScapyThread(self.serverMode)
    elif not self.soscapy.isAlive():
      self.soscapy.start()
    # ptrace ssh
    if self.session_state_addr is None:
      self.ciphers,self.session_state_addr=self.findActiveKeys(maxNum = self.maxNum)
    else:
      self.ciphers,self.session_state_addr=self.findActiveKeys(offset=self.session_state_addr)
    if self.ciphers is None:
      raise ValueError('Struct not found')
    # unstop() the process
    ### Forked no useful self.process.cont()
    # process is running... sniffer is listening
    log.info('Please make some ssh  traffic')  
    self.init()
    self.loop()
    log.info("done for pid %d, struct at 0x%lx"%(self.process.pid, self.session_state_addr))
    return
    
  def init(self):
    ''' plug sockets, packetizer and outputs together '''
    receiveCtx,sendCtx = self.ciphers.getCiphers()
    # Inbound
    log.info('activate INBOUND receive')
    self.inbound['context'] = receiveCtx
    self.inbound['socket'] = self.soscapy.getInboundSocket()
    self.inbound['packetizer'] = Packetizer( self.inbound['socket'] )
    self.inbound['packetizer'].set_log(logging.getLogger('inbound.packetizer'))
    self.inbound['engine'] = self.activate_cipher(self.inbound['packetizer'], receiveCtx )
    self.inbound['filewriter'] =  output.SSHStreamToFile(self.inbound['packetizer'], self.inbound, 'ssh-in')

    # out bound
    log.info('activate OUTBOUND send')
    self.outbound['context'] = sendCtx
    self.outbound['socket'] = self.soscapy.getOutboundSocket()
    self.outbound['packetizer'] = Packetizer(self.outbound['socket'])
    self.outbound['packetizer'].set_log(logging.getLogger('outbound.packetizer'))
    self.outbound['engine'] = self.activate_cipher(self.outbound['packetizer'], self.outbound['context'] )
    self.outbound['filewriter'] =  output.SSHStreamToFile(self.outbound['packetizer'], self.outbound, 'ssh-out')

    self.worker = output.Supervisor()
    self.worker.add( self.inbound['socket'], self.inbound['filewriter'].process )
    self.worker.add(self.outbound['socket'], self.outbound['filewriter'].process )
    return
    
  def loop(self):
    self.worker.run()
    return

  def activate_cipher(self, packetizer, context):
    "switch on newly negotiated encryption parameters for inbound traffic"
    #packetizer.set_log(log)
    #packetizer.set_hexdump(True)
    # find Engine from engine.ciphers
    engine = CIPHERS[context.name](context) 
    log.debug( 'cipher:%s block_size: %d key_len: %d '%(context.name, context.block_size, context.key_len ) )
    #print engine, type(engine)
    mac = context.mac
    if mac is not None:
      mac_key    = mac.key #XXX TODO
      mac_engine = Transport._mac_info[mac.name]['class']
      mac_len = mac.mac_len
    # again , we need a stateful HMAC engine. 
    # we disable HMAC checking to get around.
    # fix our engines in packetizer
    ## packetizer.set_hexdump(True)
    packetizer.set_inbound_cipher(engine, context.block_size, mac_engine, mac_len , mac_key)
    if context.comp.enabled != 0:
      name = context.comp.name
      compress_in = Transport._compression_info[name][1]
      log.debug('Switching on inbound compression ...')
      packetizer.set_inbound_compressor(compress_in())
    #ok
    return engine
  
  def refresh(self):
    log.warning('Refreshing Engine states')
    # ptrace ssh
    if self.session_state_addr is None:
      self.ciphers,self.session_state_addr=self.findActiveKeys(maxNum = self.maxNum)
    else:
      self.ciphers,self.session_state_addr=self.findActiveKeys(offset=self.session_state_addr)
    if self.ciphers is None:
      raise ValueError('Struct not found')
    # refresh both engine
    receiveCtx,sendCtx = self.ciphers.getCiphers()
    # Inbound
    log.info('activate INBOUND receive')
    self.inbound['context'] = receiveCtx
    self.outbound['context'] = sendCtx
    self.inbound['engine'].sync(self.inbound['context'])
    self.outbound['engine'].sync(self.outbound['context'])
    pass



def launchScapyThread(serverMode):
  from threading import Thread
  # @ at param
  port=22
  sshfilter="tcp and port %d"%(port)
  #
  if serverMode:
    soscapy=socket_scapy.socket_scapy(sshfilter, isInboundPacketCallback=socket_scapy.isdestport22,
                                      isOutboundPacketCallback=socket_scapy.isNotdestport22)
    log.info(' --- SSHD SERVER MODE ---- ')
  else:
    soscapy=socket_scapy.socket_scapy(sshfilter)
    log.info(' --- SSH  CLIENT MODE ---- ')
  sniffer = Thread(target=soscapy.run)
  worker  = Thread(target=soscapy.run2)
  sniffer.worker = worker
  soscapy.setThread( sniffer )
  sniffer.start()
  worker.start()
  return soscapy


def parseSSHClient(proc,pcapfilter): ## TODO replace by search
  parser=argparser()
  opts = parser.parse_args([str(proc.pid)])
  opts.func(opts)
  #keysFinder=OpenSSHKeysFinder(proc.pid)
  #ciphers,addr=keysFinder.findActiveKeys()
  #keysFinder.save(ciphers.session_state)
  return 
  
def parseSSHServer(proc,pcapfilter):  ## TODO replace by search --server
  parser=argparser()
  opts = parser.parse_args([str(proc.pid),'--server'])
  opts.func(opts)
  #keysFinder=OpenSSHKeysFinder(proc.pid)
  #ciphers,addr=keysFinder.findActiveKeys()
  #keysFinder.save(ciphers.session_state)
  return 
  
def parseSSHAgent(proc,ignore):
  from openssl import OpenSSLStructFinder
  keysFinder=OpenSSLStructFinder(proc.pid)
  return keysFinder.findAndSave()

def usage(parser):
  parser.print_help()
  sys.exit(-1)


def argparser():
  parser = argparse.ArgumentParser(prog='sshsnoop', description='Live decription of Openssh traffic.')
  parser.add_argument('pid', type=int, help='Target PID')
  parser.add_argument('--addr', type=str, help='active_context memory address')
  parser.add_argument('--server', dest='isServer', action='store_const', const=True, help='Use sshd server mode')
  parser.set_defaults(func=search)
  return parser

def search(args):
  pid=int(args.pid)
  sessionStateAddr=None
  log.info("Target has pid %d"%pid)
  if args.addr != None:
    sessionStateAddr=int(args.addr,16)
  serverMode=args.isServer # True or None

  #logging.getLogger('model').setLevel(logging.INFO)
  ###engine,key=openssh.test(16634)
  #engine,key=testEncDec(pid)
  #return 0
  
  decryptatator=OpenSSHLiveDecryptatator(pid, sessionStateAddr=sessionStateAddr, serverMode=serverMode)
  decryptatator.run()
  log.info("done for pid %d, struct at 0x%lx"%(pid,decryptatator.session_state_addr))
  sys.exit(0)
  return -1


def main(argv):
  logging.basicConfig(level=logging.INFO)
  logging.getLogger('abouchet').setLevel(logging.INFO)
  logging.getLogger('model').setLevel(logging.INFO)
  ##logging.getLogger('openssh.model').setLevel(logging.INFO)
  ##logging.getLogger('scapy').setLevel(logging.ERROR)
  logging.getLogger('socket.scapy').setLevel(logging.INFO)
  logging.getLogger('engine').setLevel(logging.INFO)
  logging.getLogger('output').setLevel(logging.INFO)
  ##logging.getLogger('root').setLevel(logging.DEBUG)
  ##logging.getLogger('sslnoop.openssh').setLevel(logging.DEBUG)
  logging.getLogger("inbound.packetizer").setLevel(logging.INFO)
  logging.getLogger("outbound.packetizer").setLevel(logging.INFO)

  logging.debug(argv)
  # we must have big privileges...
  if os.getuid() + os.geteuid() != 0:
    log.error("You must be root/using sudo to read memory and sniff traffic.")
    return

  #sys.path.append('/home/jal/Compil/python-haystack/')
  #sys.path.append('/home/jal/Compil/sslsnoop/sslsnoop/')
  
  parser = argparser()
  opts = parser.parse_args(argv)
  try:
    opts.func(opts)
  except ImportError,e:
    log.error('Struct type does not exists.')
    print e

  #0xb9116268

  return 0
  



if __name__ == "__main__":
  main(sys.argv[1:])

#logging.basicConfig(level=logging.INFO)

#pid=31833
#addr=0xb904b268
#dec=OpenSSHLiveDecryptatator(pid,sessionStateAddr=addr)




def testEncDec(pid):
  logging.basicConfig(level=logging.INFO)
  finder=OpenSSHLiveDecryptatator(pid)
  
  ciphers,addr=finder.findActiveKeys(offset=finder.session_state_addr)
  logging.basicConfig(level=logging.DEBUG)
  engine = StatefulAES_Ctr_Engine(ciphers.receiveCtx)
  engine2 = StatefulAES_Ctr_Engine(ciphers.receiveCtx)
  
  app_data=ciphers.receiveCtx.app_data
  key,rounds=app_data.getCtx()
  print "key=",repr(key)
  print len(key)
  print "rounds=",rounds
  print ''
  
  print "waiting for packet:"
  buf=b'1234567890ABCDEF1234567890ABCDEF'
  
  print '"aes_counter" : "%s"'%repr(engine.aes_key_ctx.getCounter())
  #encrypted=engine.decrypt(buf)
  encrypted=engine.decrypt(buf)
  print 'Encrypted len', len(encrypted)
  decrypted=engine2.decrypt(encrypted)
  print 'engine 1 "aes_counter" : "%s"'%repr(engine.aes_key_ctx.getCounter())
  print 'engine 2 "aes_counter" : "%s"'%repr(engine2.aes_key_ctx.getCounter())
  print 'decrypted=',decrypted
  return engine,key

def test(pid,addr):
  logging.basicConfig(level=logging.INFO)
  soscapy=launchScapyThread()
  finder=OpenSSHLiveDecryptatator(pid,sessionStateAddr=addr)

  ciphers,addr=finder.findActiveKeys(offset=finder.session_state_addr)
  logging.basicConfig(level=logging.DEBUG)
  finder.process.cont()

  #readso=soscapy.getInboundSocket()
  #engine = StatefulAES_Ctr_Engine(ciphers.receiveCtx)
  #app_data=ciphers.receiveCtx.app_data
  readso=soscapy.getOutboundSocket()
  engine = StatefulAES_Ctr_Engine(ciphers.sendCtx)
  app_data=ciphers.sendCtx.app_data
  key,rounds=app_data.getCtx()
  print "key=",repr(key)
  print len(key)
  print "rounds=",rounds
  print ''
  print "waiting for packet:"
  import time,select,struct
  start=time.time()
  while( True):
    r,w,o=select.select([readso],[],[],2)
    if len(r) > 0:
      if time.time()-start > 20:
        break
    else: 
      encrypted=readso.recv(1024)
      print engine.aes_key_ctx.__dict__
      #print '"aes_counter" : "%s"'%repr(engine.aes_key_ctx.getCounter())
      #print 'Encrypted len', len(encrypted)
      print '"aes_counter" : "%s"'%repr(engine.aes_key_ctx.getCounter())
      decrypted=engine.decrypt(encrypted)
      print '"aes_counter" : "%s"'%repr(engine.aes_key_ctx.getCounter())
      print 'decrypted=',decrypted
      packet_size = struct.unpack('>I', decrypted[:4])[0]
      print 'packet_size BE',packet_size
      packet_size = struct.unpack('<I', decrypted[:4])[0]
      print 'packet_size LE',packet_size
  
  # find it again
  ciphers,addr=finder.findActiveKeys(offset=finder.session_state_addr)
  app_data=ciphers.receiveCtx.app_data
  print 'our aes_counter : "%s"'%repr(engine.aes_key.getCounter())
  print 'its aes_counter : "%s"'%repr(app_data.aes_key.getCounter())

  #print "key=",repr(key)
  #buf=readso.recv(1024)
  #decrypted=engine.decrypt(encrypted[:AES_BLOCK_SIZE])
  
  return engine,key

  

def testSimpleDecrypt(readso,engine_in):
  block = readso.recv(16)
  print socket_scapy.hexify(block)
  header = engine_in.decrypt(block)
  print 'First 4 char  data=',repr(header[:4])
  print 'p_size %d or %d'%(struct.unpack('>I', header[:4])[0], struct.unpack('<I', header[:4])[0])
  #print 'p_size %d or %d'%(struct.unpack('>I', header[:4])[3], struct.unpack('<I', header[:4])[3])
  packet_size = struct.unpack('>I', header[:4])[0]
  # leftover contains decrypted bytes from the first block (after the length field)
  leftover = header[4:]
  print 'packet_size=%d \nleftoverlen=%d'%(packet_size,len(leftover))
  if (packet_size - len(leftover)) % 16 != 0:
    print "SSHException('Invalid packet blocking')"
  print 'Test descrypt Finished\n\n\n'  
  return  



#test(2517,0xb9002268)





#test read memory

#import ptrace
#f=file('/proc/8902/maps')
#lines=f.readlines()

import logging,sys,os
logging.basicConfig(level=logging.DEBUG)

import abouchet,ctypes_openssl,ctypes_openssh
import ctypes
from ptrace.debugger.debugger import PtraceDebugger
from ptrace.debugger.memory_mapping import readProcessMappings
from ctypes import *
from ptrace.ctypes_libc import libc


log=logging.getLogger('test')

def printBytes(data):
  for i in range(0,len(data)/8,8):
    print "0x%lx"%data[i],
    

PID=27477
ADDR=0xb835b4e8


#dbg=PtraceDebugger()
#process=dbg.addProcess(pid,is_attached=False)
#if process is None:
#  log.error("Error initializing Process debugging for %d"% pid)
#  sys.exit(-1)

#maps=readProcessMappings(process)
#stack=process.findStack()




def dbg_read(addr):
  from ptrace.cpu_info import CPU_64BITS, CPU_WORD_SIZE, CPU_POWERPC
  for a in range(addr,addr+88,CPU_WORD_SIZE):
    print "0x%lx"%process.readWord(a)


def readRsa(addr):
  dbg=PtraceDebugger()
  process=dbg.addProcess(PID, is_attached=False)
  if process is None:
    log.error("Error initializing Process debugging for %d"% PID)
    sys.exit(-1)
  # read where it is
  ######################### RAAAAAAAAAAAAH
  rsa=ctypes_openssl.RSA.from_buffer_copy(process.readStruct(addr,ctypes_openssl.RSA))
  mappings=readProcessMappings(process)
  print "isValid : ", rsa.isValid(mappings)
  #rsa.printValid(maps)
  print rsa
  #print rsa.n
  #print rsa.n.contents
  #print ctypes.byref(rsa.n.contents)
  #print rsa
  print '------------ === Loading members'
  ret=rsa.loadMembers(process,mappings)
  print '------------ === Loading members finished'
  print ret,rsa
  return rsa

def readDsa(addr):
  dbg=PtraceDebugger()
  process=dbg.addProcess(PID, is_attached=False)
  if process is None:
    log.error("Error initializing Process debugging for %d"% pid)
    sys.exit(-1)
  # read where it is
  dsa=ctypes_openssl.DSA.from_buffer_copy(process.readStruct(addr,ctypes_openssl.DSA))
  mappings=readProcessMappings(process)
  print "isValid : ", dsa.isValid(mappings)
  #dsa.printValid(maps)
  #print 'DSA1 -> ', dsa
  #print '------------'
  #print 'DSA1.q -> ', dsa.q
  print '------------ === Loading members'
  dsa.loadMembers(process,mappings)
  #print '------------  ===== ==== '
  #print 'DSA2.q -> ', dsa.q
  #print 'DSA2.q.contents -> ', dsa.q.contents
  #print ctypes.byref(rsa.n.contents)
  #print dsa
  return dsa

def writeWithLibRSA(addr):
  ssl=cdll.LoadLibrary("libssl.so")
  # need original data struct
  #rsa=process.readBytes(addr, ctypes.sizeof(ctypes_openssl.RSA) )
  #rsa=ctypes.addressof(process.readStruct(addr,ctypes_openssl.RSA))
  rsa=readRsa(addr)
  rsa_p=ctypes.addressof(rsa)
  print 'rsa acquired 0x%lx copied to 0x%lx'%(addr,rsa_p)
  f=libc.fopen("test.out","w")
  print 'file opened',f  
  ret=ssl.PEM_write_RSAPrivateKey(f, rsa_p, None, None, 0, None, None)
  print 'key written'  
  print ret,f

def writeWithLibDSA(addr):
  ssl=cdll.LoadLibrary("libssl.so")
  dsa=readDsa(addr)
  dsa_p=ctypes.addressof(dsa)
  print 'dsa acquired 0x%lx copied to 0x%lx'%(addr,dsa_p)
  f=libc.fopen("test.out","w")
  print 'file opened',f  
  ret=ssl.PEM_write_DSAPrivateKey(f, dsa_p, None, None, 0, None, None)
  print 'key written'  
  print ret,f

def withM2(addr):
  import M2Crypto
  from M2Crypto.BIO import MemoryBuffer
  from M2Crypto import RSA as mRSA
  rsa=process.readBytes(addr, ctypes.sizeof(ctypes_openssl.RSA) )
  bio=MemoryBuffer(rsa)
  # tsssi need PEM
  myrsa=mRSA.load_key_bio(bio)
  return myrsa

def printSize():
  ctypes_openssl.printSizeof()
  ctypes_openssh.printSizeof()

def printme(obj):
  log.info(obj)

def findCipherContext():
  dbg=PtraceDebugger()
  process=dbg.addProcess(pid,is_attached=False)
  if process is None:
    log.error("Error initializing Process debugging for %d"% pid)
    sys.exit(-1)
  mappings=readProcessMappings(process)
  stack=process.findStack()
  for m in mappings:
    #if m.pathname != '[heap]':
    #  continue
    if not abouchet.hasValidPermissions(m):
      continue    
    print m,m.permissions
    abouchet.find_struct(process, m, ctypes_openssh.CipherContext, printme)

def isMemOf(addr,mpid):
  class p:
    pid=mpid
  myP=p()
  for m in readProcessMappings(myP):
    if addr in m:
      print myP, m
      return True
  return False

def testScapy():
  import socket_scapy
  socket_scapy.test()

def testScapyThread():
  import socket_scapy,select
  from threading import Thread
  port=22
  sshfilter="tcp and port %d"%(port)
  soscapy=socket_scapy.socket_scapy(sshfilter,packetCount=100)
  log.info('Please make some ssh  traffic')
  sniffer = Thread(target=soscapy.run)
  sniffer.start()
  # sniffer is strted, let's consume
  nbblocks=0
  data=''
  readso=soscapy.getInboundSocket()
  while sniffer.isAlive():
    r,w,oo=select.select([readso],[],[],1)
    if len(r)>0:
      data+=readso.recv(16)
      nbblocks+=1
  # try to finish socket
  print 'sniffer is finished'
  r,w,oo=select.select([readso],[],[],0)
  while len(r)>0:
    data+=readso.recv(16)
    nbblocks+=1
    r,w,oo=select.select([readso],[],[],0)
  #end      
  print "received %d blocks/ %d bytes"%(nbblocks,len(data))
  print 'sniffer captured : ',soscapy




#rsa=readRsa(addr)

#writeWithLibRSA(addr)
#print '---------------'
#abouchet.find_keys(process,stack)

#dsa=readDsa(0xb835b4e8)
#print dsa
#rsa=readRsa(0xb835c9a0)
#writeWithLibDSA(addr)

#printSize()

#x pourPEM_write_RSAPrivateKey

class B(ctypes.Structure):
  _fields_=[("b1",ctypes.c_ulong),('b2',ctypes.c_ulonglong)]

class A(ctypes.Structure):
  _fields_=[("a1",ctypes.c_ulong),('b',ctypes.POINTER(B)),('a2',ctypes.c_ulonglong)]

myaddr=0xb7c16884
# rsa-> n
myaddr=0xb8359148
pid=12563
if len(sys.argv) == 2:
  myaddr=int(sys.argv[1],16)
  if isMemOf(myaddr, pid ):
    print "isMemOf(0x%lx, 8831) - python"%myaddr
  if isMemOf(myaddr, PID):
    print "isMemOf(0x%lx, 27477) - ssh-agent"%myaddr

#findCipherContext()


soscapy=testScapyThread()









import scapy.config 

class a:
  def __init__(self):
    scapy.config.conf.use_pcap=True
    self.packetCount = 10
    self.timeout = 10
    self.filterRules = 'tcp and port 22'    
    return
  def enqueue(self,p):
    print len(p), p['TCP'].seq, p.summary()
  def run(self):
    from scapy.all import sniff
    print ('Using L2listen = %s'%(scapy.config.conf.L2listen)) 
    sniff(iface='any', count=self.packetCount,timeout=self.timeout,store=0,filter=self.filterRules,prn=self.enqueue)
    print ('============ SNIFF Terminated ====================')


o=a()

o.run()




# -*- coding: utf-8 -*-
from setuptools import setup
from glob import glob

setup(name="sslsnoop",
    version="0.2",
    description="Dumps the live traffic of an ssl-encrypted stream.",
    long_description=open('README').read(),

    url="http://packages.python.org/sslsnoop/",
    download_url="http://github.com/trolldbois/sslsnoop/tree/master",
    license='GPL',
    classifiers=[
        "Topic :: System :: Networking",
        "Topic :: Security",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Programming Language :: Python",
        "Development Status :: 5 - Production/Stable",
    ],
    keywords=['memory','analysis','forensics','struct','ptrace','openssh','openssl','decrypt'],
    author="Loic Jaquemet",
    author_email="loic.jaquemet+python@gmail.com",
    packages = ['sslsnoop'],
    scripts = ['scripts/sslsnoop-openssh', 'scripts/sslsnoop-openssl'],
    extras_require = {
        'haystack':  ["python-haystack"],
        'ptrace':  ["python-ptrace"],
    },
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2011 Loic Jaquemet loic.jaquemet+python@gmail.com
#

__author__ = "Loic Jaquemet loic.jaquemet+python@gmail.com"

import os,logging,psutil,sys, time
import openssh
import subprocess

log=logging.getLogger('finder')

_targets={
    'ssh': openssh.parseSSHClient,
    'sshd': openssh.parseSSHServer,
    'ssh-agent': openssh.parseSSHAgent,
    #'firefox': []
    }


def buildTuples(targets):
  rets=[]
  for proc in psutil.process_iter():
    if proc.name in targets:
      rets.append( (proc.pid,proc) )
  return rets

    
def pgrep(name):
  pids=[]
  for proc in psutil.process_iter():
    if proc.name == name:
      pids.append(name)
  return pids

def makeFilter(conn):
  # [connection(fd=3, family=2, type=1, local_address=('192.168.1.101', 36386), remote_address=('213.186.33.2', 22), status='ESTABLISHED')]
  pcap_filter = "host %s and port %s and host %s and port %s" %(conn.local_address[0],conn.local_address[1] ,
                            conn.remote_address[0],conn.remote_address[1]  )
  return pcap_filter

def checkConnections(proc):
  conns=proc.get_connections()
  if len(conns) == 0 :
    return False
  elif len(conns) > 1 :
    if proc.name == 'sshd':
      estab = [ c for c in conns if c.status == 'ESTABLISHED'] 
      log.info('Found %d connections for %s'%(len(estab), proc.name))
      return ' or '.join(makeFilter(estab))
    log.warning(' %s has more than 1 connection ?'%(proc.name))
    return False
  elif conns[0].status != 'ESTABLISHED' :
    log.warning(' %s has noESTABLISHED connections (1 %s)'%(conn[0].status))
    return False
  # else ok for ssh
  f=makeFilter(conns[0])
  log.info('Found connection %s for %s'%(conns[0], proc.name))
  return f

def usage(txt):
  log.error("Usage : %s "% txt)
  sys.exit(-1)


def main(argv):
  logging.basicConfig(level=logging.INFO)
  #logging.getLogger('model').setLevel(logging.INFO)


  # we must have big privileges...
  if os.getuid() + os.geteuid() != 0:
    log.error("You must be root/using sudo to read memory and sniff traffic. So there's no point in going further")
    return
  
  if not os.access('outputs', os.X_OK) :
    os.mkdir('outputs/')
  
  options=buildTuples(_targets)
  threads=[]
  forked=0
  for pid,proc in options:
    log.info("Searching in %s/%d memory"%(proc.name,proc.pid))
    pcap_filter = checkConnections(proc)
    if not pcap_filter and not 'ssh-agent' == proc.name:
      continue
    # call cb
    if (os.fork() == 0):
      _targets[proc.name](proc, pcap_filter)
      sys.exit(0)
      break
    forked+=1
    log.info('Subprocess launched on pid %d'%(proc.pid))

  time.sleep(5)
  log.info(' ============== %d process forked. look into outputs/ for data '%(forked))
  sys.exit(0)
  return 0


if __name__ == "__main__":
  main(sys.argv[1:])


  
  
  
